import logging

from typing import List, Dict, Tuple, Any

import dominate
from dominate.tags import h4, table, tr, td, img

from ppa import ui
from ppa_tools import status_icons, vendor_icons

logger = logging.getLogger(f"ppa-tools.{__name__}")

VENDOR_MAP = {
    "ppa": vendor_icons.PPA,
    "pxm": vendor_icons.OSIRIUM_O,
    "ubuntu": vendor_icons.UBUNTU,
    "vault": vendor_icons.VAULT,
    "active_directory": vendor_icons.ACTIVE_DIRECTORY,
    "service_now": vendor_icons.SERVICE_NOW,
    "infoblox": vendor_icons.INFOBLOX,
    "cisco": vendor_icons.CISCO,
    "windows_server": vendor_icons.WINDOWS_SERVER,
    "workbooks": vendor_icons.WORKBOOKS,
    "freshdesk": vendor_icons.FRESHDESK,
    "free_pbx": vendor_icons.FREE_PBX,
    "email": vendor_icons.EMAIL,
    "ansible": vendor_icons.ANSIBLE,
    "aws": vendor_icons.AWS,
    "azure": vendor_icons.AZURE,
    "github": vendor_icons.GITHUB,
    "vsphere": vendor_icons.VSPHERE,
    "office_365": vendor_icons.OFFICE_365,
    "exchange": vendor_icons.EXCHANGE_MINI,
    "twilio": vendor_icons.TWILIO_MINI,
    "checkpoint": vendor_icons.CHECKPOINT_MINI,
}

INVERTED_ICON_MAP = {
    "queued": None,
    "succeeded": status_icons.TICK_INVERTED,
    "failed": status_icons.CROSS_INVERTED,
    "skipped": status_icons.SKIPPED_INVERTED,
    "in_progress": status_icons.IN_PROGRESS_INVERTED,
}


class TaskSteps:
    def __init__(self, steps: List[Tuple[str, str]], title: str = "Task Steps", output_id: str = "task_steps") -> None:
        """ Create a task progress table with defined steps. Used to track task
        progress and display it back to the PPA user. Steps can be assigned vendor
        icons. The status of each step can be assigned using the available class
        methods, and the table can be displayed to the PPA user at any point.

        :param steps: A list of task step 2-tuples, where the first element the
            name, and the second element is a vendor icon ID or left blank.
        :param title: The title of the table when displayed in the PPA interface.
        :param output_id: The element ID of the displayed table.

        :raises TypeError: if the value of steps is not a list.

        Example:

            .. code-block:: python

                steps = TaskSteps(
                    [
                        ("Get Hashicorp Vault Secrets", "vault"),
                        ("Get Active Directory User", "active_directory"),
                        ("Select Operations", "ppa"),
                        ("Perform Operations", "active_directory"),
                    ],
                    title="Task Progress",
                    output_id="task_progress",
                )

        .. seealso:: :ref:`display:Creating & Displaying a Task Steps Table`

        """
        if not isinstance(steps, list):
            message: str = "Task steps must be in list format."
            ui.output_error(message)
            raise TypeError(message)

        self.title: str = title
        self.id: str = output_id

        # Create a list of dictionaries
        # in which we store the name and
        # current status of each step.
        self.steps: List[Dict[str, Any]] = []
        for step in steps:
            if not isinstance(step, tuple):
                step = (step, None)
            step_name, vendor_name = step
            self.steps.append(
                {
                    "name": step_name,
                    "status": "queued",
                    "vendor": vendor_name,
                    "vendor_icon": VENDOR_MAP.get(vendor_name),
                }
            )

        # Keep track of what the next step is
        self.current_step_index: int = 0

    def show_all(self) -> None:
        """ Display the steps table in the PPA interface. Calling this function builds
        an HTML table using the current statuses of all steps, and embeds it in a markdown
        document. The document is displayed by the `output_markdown` function from the `ppa`
        library.
        """
        document = dominate.document("")
        # Open the document header and add an h3
        with document.head:
            h4(self.title)
        # Open the document, defaulting to 'body'
        with document:
            # Create the table and keep it open so we can populate it.
            with table(style="border-spacing: 0;"):
                # For every task step, create a table row and two table columns.
                for step in self.steps:
                    name = step["name"]
                    status = step["status"]
                    status_icon = INVERTED_ICON_MAP[status]
                    if step["vendor_icon"]:
                        vendor_icon = step["vendor_icon"]["base_64"]
                        vendor_icon_format = step["vendor_icon"]["format"]
                    with tr(style=f"border: none; background: transparent;", height=5):
                        # Add the vendor column
                        if step["vendor_icon"]:
                            td(style=f"border: none; vertical-align: middle;").add(
                                img(
                                    src=f"data:image/{vendor_icon_format};base64,{vendor_icon}",
                                    style=f"background: transparent; border: none; vertical-align: middle;",
                                    alt=step["vendor"] if step["vendor"] else "",
                                    width="20",
                                    height="20",
                                )
                            )
                        else:
                            td(style=f"border: none; vertical-align: middle;")
                        # Add the name column
                        if status == "in_progress":
                            font_weight = "bold"
                            font_style = "regular"
                        elif status == "failed":
                            font_weight = "bold"
                            font_style = "italic"
                        else:
                            font_weight = "regular"
                            font_style = "regular"
                        td(
                            name,
                            style=(
                                f"border: none; font-size: 14px; font-weight: {font_weight}; "
                                f"font-style: {font_style}; vertical-align: middle; padding-top: 10px;"
                            ),
                        )
                        # Add the status column
                        with td(style=(f"border: none; vertical-align: middle;")):
                            if status_icon:
                                td.add(
                                    img(
                                        src=f"data:image/{status_icon['format']};base64,{status_icon['base_64']}",
                                        style="background: transparent; vertical-align: middle;",
                                        alt=step["status"],
                                        width="20",
                                        height="20",
                                    )
                                )

        document = document.render().replace("\\n", "")
        ui.output_markdown(document, element_id=self.id)

    def _update_step(self, index: int, status: str) -> None:
        """ Update a self.steps dictionary 'status' key.

            index: int. The index of the step to update.
            status: str. Must exist in INVERTED_ICON_MAP.
        """
        try:
            assert status in INVERTED_ICON_MAP
            logger.info(f'updating step {index} from {self.steps[index]["status"]} to {status}')
            self.steps[index]["status"] = status
        except AssertionError:
            message = "Tried to update a task step to unknown " f"status {status}."
            ui.output_error(message)
        except IndexError:
            raise IndexError(
                "Tried to update non-existent task step index "
                f"{index} to {status}. Highest index is {len(self.steps) - 1}."
            )

    def _succeed_step(self, index: int) -> None:
        """
        Args:
            index: int. The index of the step to update.
        """
        self._update_step(index, "succeeded")

    def _fail_step(self, index: int) -> None:
        """ Mark a step as failed by appending the
        cross image to the end of the step string.

        Args:
            index: int. The index of the step to update.
        """
        self._update_step(index, "failed")

    def _skip_step(self, index: int) -> None:
        self._update_step(index, "skipped")

    def _step_in_progress(self, index: int) -> None:
        self._update_step(index, "in_progress")

    def succeed_next(self):
        """
        - Set the next unresolved step to `succeeded`.
        - Increment the current step index.
        """
        self._succeed_step(self.current_step_index)
        self.current_step_index += 1

    def skip_next(self):
        """
        - Set the next unresolved step to `skipped`.
        - Increment the current step index.
        """
        self._skip_step(self.current_step_index)
        self.current_step_index += 1

    def fail_next(self):
        """
        - Set the next unresolved step to `failed`.
        - Increment the current step index.
        """
        self._fail_step(self.current_step_index)
        self.current_step_index += 1

    def next_in_progress(self):
        """
        - Set the next unresolved step to `in progress`.

        *Does not increment the current step index.*
        """
        self._step_in_progress(self.current_step_index)

    def succeed_remaining(self):
        """
        - Set **all** unresolved steps to `succeeded`.
        - Exhaust the current step index.
        """
        for _ in self.steps[self.current_step_index :]:
            self.succeed_next()

    def fail_remaining(self):
        """
        - Set **all** unresolved steps to `failed`.
        - Exhaust the current step index.
        """
        for _ in self.steps[self.current_step_index :]:
            self.fail_next()

    def skip_remaining(self):
        """
        - Set **all** unresolved steps to `skipped`.
        - Exhaust the current step index.
        """
        for _ in self.steps[self.current_step_index :]:
            self.skip_next()

    def task_started(self):
        """
        - Set the next unresolved step to `in progress`.
        - Display the table with `show_all()`.

        *Does not increment the current step index.*
        """
        self.next_in_progress()
        self.show_all()

    def proceed(self):
        """
        - Set the next unresolved step to `succeeded`.
        - Increment the current step index.
        - Set the subsequent step to `in progress`.
        - Display the table with `show_all()`.
        """
        self.succeed_next()
        self.next_in_progress()
        self.show_all()

    def task_failed(self):
        """
        - Set the next unresolved step to `failed`.
        - Set **all** subsequent steps to `skipped`.
        - Display the table with `show_all()`.
        - Exhaust the current step index.
        """
        self.fail_next()
        self.skip_remaining()
        self.show_all()

    def task_finished(self):
        """
        - Set the next unresolved step to `succeeded`.
        - Display the table with `show_all()`.
        - Exhaust the current step index.
        """
        self.succeed_next()
        self.show_all()

    def rollback(self, new_current_step: int, status: str = "queued"):
        """ Roll the steps back to a new index. This can be used when
        a user needs to start over from an earlier point inside an PPA
        task.

        :param new_current_step: The index to set as the new current step index.
        :param status: The status to set against all previously resolved steps
            that occur after the `new_current_step`.

        :raises IndexError: if the supplied `new_current_step` index does not exist in the TaskSteps `steps` list.

        Example:

        - Roll back to the first step, setting all previously resolved steps to `queued`.
        - Set the first step to `in progress` and display the table with :py:func:`task_started`.

        .. code-block:: python

            steps.rollback(0)
            steps.task_started()

        .. seealso:: :ref:`display:Creating & Displaying a Task Steps Table`, :ref:`display:Current Step Index`

        """
        logger.info(f"rolling back to step {new_current_step}")
        rollback_step = self.current_step_index
        while rollback_step >= new_current_step:
            logger.info(f"rolling step {rollback_step} back to {status}")
            try:
                self._update_step(rollback_step, status)
            except IndexError:
                raise IndexError(
                    f"Tried to rollback non-existent step {rollback_step}, valid steps are 0-{len(self.steps) - 1}"
                )
            rollback_step -= 1
        self.current_step_index = new_current_step
