"""
Generate

This module contains actions used to generate various types of data.
"""

import uuid

from typing import List, Dict, Any, Union

from ppa import ui
from ppa_tools import generate


def table_from_dictionaries(
    header: List[str], keys: List[str], dictionaries: List[Dict[str, Any]]
) -> Dict[str, Union[List[str], List[List[str]]]]:
    r"""Generate table data from a list of dictionaries.

!!! question "Minimum Plugin Version: 1.3.0"

??? info "Input"

    - __header:__ a list of table column headers

    - __keys:__ a list of corresponding dictionary keys, one for each `header`

    - __dictionaries:__ a list of dictionaries to get the values of each `key` from

??? success "Output"

    A dictionary containing the keys `header` & `rows`.

??? example

    - Finding failed SSH logons using [extract_values](../text/#extract_values) & saving them as `failed_logons`

    - Using this action to turn `failed_logons` into table data, & saving it as `failed_logons_table`

    - Displaying a table using the data generated by this action

    ```YAML linenums="1" hl_lines="10 20 21 26 27"
    - ppa_tools.text.extract_values:
        extract:
          - name: username
            expression: for (.*) from
          - name: ip_address
            expression: from (\d+\.\d+\.\d+\.\d+) port
        text: >
          2020-08-11T08:19:40.463248+00:00 machine-name sshd[10349]: Failed password for user_1 from 1.2.3.4 port 58041 ssh2
          2020-08-12T13:24:07.463248+00:00 machine-name sshd[10349]: Failed password for user_2 from 2.3.4.5 port 58041 ssh2
      save: failed_logons

    - ppa_tools.generate.table_from_dictionaries:
        header:
          - Username
          - IP Address
        keys:
          - username
          - ip_address
      load:
        dictionaries: failed_logons
      save: failed_logons_table

    - ppa.ui.output_table:
        text: Failed SSH Logons
      load:
        header: failed_logons_table.header
        rows: failed_logons_table.rows
    ```

!!! tip "Headers & Keys"

    The `header` & `keys` lists must be the same length, & the attributes must be in the same order.

    If a supplied key is missing from a dictionary, its value will be blank.
    """
    if len(header) != len(keys):
        message = "Cannot generate table data as the supplied 'header' & 'keys' lists are not the same length."
        ui.output_error(message)
        raise ValueError(message)
    return {"header": header, "rows": [[d.get(key) or "" for key in keys] for d in dictionaries]}


def html_table(header: List[str], rows: List[List[str]]) -> str:
    r"""Generate an HTML table to use in a markdown document or email.

!!! question "Minimum Plugin Version: 1.2.0"

??? info "Input"

    - __header:__ a list of table column headers

    - __rows:__ the table contents as a list of lists, where each inner list is a row of values

??? success "Output"

    The generated HTML table as a string.

??? example

    ```YAML linenums="1"
    - ppa_tools.html.generate_table:
        header:
          - City
          - Country
        rows:
          - - Paris
            - France
          - - London
            - England
      save: html_table
    ```
    """
    header_styles = "; ".join(
        ["padding: 5px", "font-size: 13px", "font-weight: heavy", "text-align: center", "border: 1px solid grey"]
    )
    cell_styles = "; ".join(["padding: 5px", "font-size: 12px", "border: 1px solid gray", "text-align: center"])

    row_template = "<tr>\n{row}\n</tr>"
    table_header = row_template.format(row="\n".join([f'<th style="{header_styles}">{head}</th>' for head in header]))
    table_rows = "\n".join(
        [row_template.format(row="\n".join([f'<td style="{cell_styles}">{cell}</td>' for cell in row])) for row in rows]
    )

    return f"""<table style="margin: auto;">
    {table_header}
    {table_rows}
</table>"""


def regex_password(regex: str = r"[A-Za-z0-9\-£\:\$\#\%\&\^]", length: int = 24) -> str:
    r"""Generate a password based on a regular expression.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __regex:__ a regular expression containing a set of characters (defaults to _[A-Za-z0-9\-£\:\$\#\%\&\^]_)

        - __length:__ the length of the generated password (defaults to _15_)

    ??? success "Output"
        The generated password as a string.

    ??? example

        Generating a 20 character password using the default character set.

        ```YAML linenums="1"
        - ppa_tools.generate.regex_password:
            length: 20
          save: generated_password
        ```
    """
    return generate.password_from_regex(regex=regex, length=length)


def dictionary_password(*, length: int = 15, word_count: int = 3, valid_word_lengths: List[int] = list(range(3, 9))) -> str:
    """Generate a password containing 1 or more dictionary words.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __length:__ the length of the generated password (defaults to _15_)

        - __word_count:__ the number of words the password should consist of (defaults to _3_)

        - __valid_word_lengths:__ a list of valid word lengths (defaults to _3_-_8_)

    ??? success "Output"
        The generated password as a string.

    ??? example

        Generating a 20 character password consisting of 4 words.

        ```YAML linenums="1"
        - ppa_tools.generate.dictionary_password:
            length: 20
            word_count: 4
          save: generated_password
        ```
    """
    return generate.password_from_words(length=length, word_count=word_count, valid_word_lengths=valid_word_lengths).password


def dictionary_password_interactive(
    *, length: int = 15, word_count: int = 3, valid_word_lengths: List[int] = list(range(3, 9))
) -> str:
    """Interactively generate a password containing 1 or more dictionary words.

    Each time a password is generated, the Task Operator will be given the choice to:

    - _Use Password_
    - _Generate Another_

    The process can be repeated as many times as needed.

    !!! question "Minimum Plugin Version: 1.0.0"

    !!! danger "Remember"

        Each generated password will be displayed to the Task Operator in a markdown document.

    ??? info "Input"

        - __length:__ the length of the generated password (defaults to _15_)

        - __word_count:__ the number of words the password should consist of (defaults to _3_)

        - __valid_word_lengths:__ a list of valid word lengths (defaults to _3_-_8_)

    ??? success "Output"
        The generated password as a string.

    ??? example

        Starting an interactive password generation process using default values.

        ```YAML linenums="1"
        - ppa_tools.generate.dictionary_password_interactive:
          save: generated_password
        ```
    """
    markdown_id = str(uuid.uuid4())
    confirm_id = str(uuid.uuid4())
    while True:
        password = generate.password_from_words(length=length, word_count=word_count, valid_word_lengths=valid_word_lengths)
        title_cased_display = " ".join([word.title() for word in password.display.split(" ")])
        ui.output_markdown(
            f"""
<html>
    <body>
        <div style="font-weight: bold; float: left;">Password: </div>
        <div style="padding-left: 5px; font-style: bold; float: left;">{password.password}</div>
        <div style="float: left; padding-left: 10px;">(</div>
            <div style="color: orange; font-style: italic; float: left;">{title_cased_display}</div>
        <div style="float: left;">)</div>
    </body>
</html>
        """,
            element_id=markdown_id,
        )
        if ui.input_confirm("Please choose an option", confirm="Use Password", cancel="Generate Another", element_id=confirm_id):
            return password.password
