import contextlib
import functools

from typing import Dict, List, Union, Optional, Callable

from ppa_tools.display import TaskSteps

StepsTable = Dict[str, Union[str, int, List[Dict[str, Union[str, Dict[str, str]]]]]]


def _as_json_table(func: Callable):
    @functools.wraps(func)
    def _inner(*args, **kwargs):
        return vars(func(*args, **kwargs))

    return _inner


@contextlib.contextmanager
def _loaded_table(table_json: StepsTable) -> contextlib._GeneratorContextManager:
    # Take the JSON representation of a TaskSteps table (generated using vars(table)) and
    # load it into a TaskSteps object, so instance methods like `proceed` and `show_all`
    # can be called on it.
    try:
        # Create a dummy empty table then set its instance attributes manually. It's
        # easier than unpacking & casting the JSON, then calling the __init__.
        new_table = TaskSteps(([]))
        for key, value in table_json.items():
            setattr(new_table, key, value)
        yield new_table
    except Exception as e:
        raise e


def _table_function(table_json: StepsTable, function_name: str) -> TaskSteps:
    with _loaded_table(table_json) as steps_table:
        getattr(steps_table, function_name)()
    return steps_table


def create(steps: List[Dict[str, str]], title: Optional[str] = "Task Progress") -> StepsTable:
    return vars(TaskSteps(list((step_dict.get("name"), step_dict.get("icon")) for step_dict in steps), title=title))


def display(table: StepsTable) -> None:
    _table_function(table, "show_all")


@_as_json_table
def start(table: StepsTable) -> StepsTable:
    return _table_function(table, "task_started")


@_as_json_table
def next_in_progress(table: StepsTable) -> StepsTable:
    return _table_function(table, "next_in_progress")


@_as_json_table
def proceed(table: StepsTable) -> StepsTable:
    return _table_function(table, "proceed")


@_as_json_table
def succeed_next(table: StepsTable) -> StepsTable:
    return _table_function(table, "succeed_next")


@_as_json_table
def succeed_remaining(table: StepsTable) -> StepsTable:
    return _table_function(table, "proceed")


@_as_json_table
def skip_next(table: StepsTable) -> StepsTable:
    return _table_function(table, "skip_next")


@_as_json_table
def skip_remaining(table: StepsTable) -> StepsTable:
    return _table_function(table, "skip_remaining")


@_as_json_table
def fail_remaining(table: StepsTable) -> StepsTable:
    return _table_function(table, "fail_remaining")


@_as_json_table
def finish(table: StepsTable) -> StepsTable:
    return _table_function(table, "task_finished")
