"""
Text

This module contains actions used to process & search text.
"""

import re
import logging

from typing import List, Dict, TypedDict, Union

from ppa import ui


logger = logging.getLogger(__name__)


class InvalidExpression(ValueError):
    pass


class AmbiguousExpression(ValueError):
    pass


class NamedGroup(TypedDict):
    r"""
- Contains a name & regular expression used to extract data from some text

- Used as an _input_ to [extract_values](../ppa_tools.text/#extract_values)

- The example below shows the dictionary structure in YAML

- See the ___required___ keys underneath the example

!!! example

    ```yaml linenums="1"
    name: username
    value: your username is (\w+)
    ```

!!! danger "Required Keys"

    __name__: The name to give the extracted value.

    __expression__: A regular expression used to find the value in the text.

    """

    name: str
    expression: str


def extract_values(values: Union[List[NamedGroup], NamedGroup], text: str) -> List[Dict[str, str]]:
    """Find & extract one or more named values from some text using regular expressions.

    This action supports both single & multiple lines of text.

    !!! question "Minimum Plugin Version: 1.3.0"

    !!! tip "Regular Expressions"

        The following [regex cheat-sheet](https://cheatography.com/davechild/cheat-sheets/regular-expressions/) & [visualiser](https://regexper.com/) can help you get started with expressions.

    ??? info "Input"

        - __values:__ a single or list of [NamedGroup](../inputs_outputs/#namedgroup) dictionaries

        - __text:__ the text to find & extract values from

    ??? success "Output"

        A list of dictionaries containing any matches found in the text.

        Each dictionary's _key_ is the `name` of the [NamedGroup](../inputs_outputs/#namedgroup), & its value is the matching text.

    ??? example

        Finding all usernames from a failed SSH logon message & saving the matches as `usernames`.

        ```YAML linenums="1"
        - ppa_tools.text.extract_values:
            values:
              name: username
              expression: password for (.*) from
            text: >
              2020-08-11T08:19:40.463248+00:00 machine-name sshd[10349]: Failed password for user_1 from 1.2.3.4 port 58041 ssh2
              2020-08-12T13:24:07.463248+00:00 machine-name sshd[10349]: Failed password for user_2 from 2.3.4.5 port 58041 ssh2
          save: usernames
        ```

        The contents of the new `usernames` variable will be:

        ```YAML
        - username: user_1
          username: user_2
        ```
    """

    def _search(text, expression):
        if match := re.search(expression, text):
            groups = match.groups()
            if len(groups) > 1:
                raise AmbiguousExpression(f"Expression '{expression}' found more than one match on line '{line}'.")
            return groups[0]
        logger.debug(f"no match for '{expression}' in '{repr(line)}'")
        return None

    values = [values] if not isinstance(values, list) else values

    invalid_expressions = []
    for named_group in values:
        try:
            re.compile(named_group["expression"])
        except re.error:
            invalid_expressions.append(named_group["expression"])
    if invalid_expressions:
        ui.output_error("One or more expressions supplied to search_text is invalid. Please view the logs for more information.")
        raise InvalidExpression("The following expressions are invalid: {}".format(", ".join(invalid_expressions)))

    matches = []
    for line in [line for line in text.split("\n") if line]:
        matches_on_line = {}
        for named_group in values:
            if result := _search(line, named_group["expression"]):
                matches_on_line[named_group["name"]] = result
        if not matches_on_line:
            logger.info(f"no matches on line '{line}'")
            # Ignore the line if no matches were found.
            continue
        matches.append(matches_on_line)
    return matches
